// -*- C++ -*-
// $RCSfile: pixelsetboundary.h,v $
// $Revision: 1.18 $
// $Author: vrc $
// $Date: 2007-09-26 22:02:14 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modifed
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef PIXELSETBOUNDARY_H
#define PIXELSETBOUNDARY_H

#include <map>
#include <vector>

#include "common/coord.h"
#include "common/geometry.h"

class CMicrostructure;

// Special classes for keeping track of the boundaries of pixel sets.

// PixelBdySegment object is either the bottom or left-side boundary
// of the pixel located at ICoord in the mesh.  The purpose of
// identifying things this way is to evade any possible round-off
// problems.  Note that, for pixel boundary segments on the right-hand
// edge or the top of a microstructure, the ICoord passed in will not
// correspond to an actual pixel in that microstructure -- the segment
// will be (e.g.) the left-hand boundary of a farther-than-rightmost
// pixel.

// The boolean flags provide enough data to locate and direct the
// segment, but this is not done initially due to possible round-off
// problems.  All these segments are either horizontal or vertical.
// The "local" flag is true if this segment is the bottom or left
// segment of the pixel indicated by the ICoord, and false if it is
// the top segment of the pixel below the ICoord or the right segment
// of the pixel to the left of the ICoord.

class PixelSetBoundary;


// Special utility class for pixel boundary intersections, so we can
// keep track of "incoming" versus "outgoing" intersections.  Incoming
// and outgoing are intended to be used with respect to elements --
// entry==true means that the pixel boundary enters the element at
// this intersection, and entry==false means that it exits the element
// at this intersection.
class PixelBdyIntersection {
private:
  // The node-index and fraction values are specific to a particular
  // skeleton element, and are filled in by the cskeleton element.
  // The node index is an integer indicating which node is at the
  // start of the element boundary on which the intersection occurs,
  // and the "fraction" number is the square of the distance from the
  // start of this element boundary to the location of the
  // intersection.
  int node_index;
  double fraction; 
public:
  Coord location;
  bool entry;  // 

  PixelBdyIntersection(Coord &loc, bool ntry) 
    : node_index(0), fraction(0), location(loc), entry(entry) {} 
  PixelBdyIntersection() : node_index(0), fraction(0), 
			   location(Coord(0.0,0.0)), entry(false) {}
  void set_element_data(int ni, double f) {
    node_index = ni;
    fraction = f;
  }
  int get_element_node_index() const { return node_index; }
  double get_element_fraction() const { return fraction; }
};

typedef std::pair<Coord, Coord> CoordPair;

class PixelBdySegment {
  // Represents a collinear and contiguous set of pixel edges.
private:
  ICoord pixel;
  bool horizontal;
  bool local;
  mutable bool coordinated; // Indicates if start, end have been set.
  mutable ICoord start, end;

  // For speed, geometric data is represented several ways.  fp_level
  // is the "level" (x-value of vertical, y-value of horizontal) of
  // this segment.  fp_start and fp_end are the start and end points
  // on the x-axis for horizontal, and y-axis for vertical, segments.
  // fp_first and fp_last are a second representation of the start and
  // end coordinates, used in intersection checking, and guaranteed to
  // have the property that fp_first < fp_last.  This saves continual
  // checking of the "local" flag.
  mutable double fp_level, fp_start, fp_end, fp_first, fp_last;
  mutable bool floated; // Indicates if the above have been set.
public:
  PixelBdySegment(const ICoord &pxl, bool horiz, bool lcl) :
    pixel(pxl), horizontal(horiz), local(lcl), 
    coordinated(false), floated(false) {}
  bool operator<(const PixelBdySegment &other) const;
  ICoord coordinate() const;
  void extend();
  void set_floats(const Coord&) const;

  // Intersection-related support functions for the skeleton elements'
  // categoryAreas computation.
  bool find_no_intersection(const Coord* const, int, const Coord&) const;

  int find_one_intersection(const Coord* const, int,
			    PixelBdyIntersection&, bool) const;

//   bool find_intersection(const Coord &p1, const Coord &p2, 
// 			 const Coord &dp,
// 			 PixelBdyIntersection &) const;

  inline Coord start_pt() const {
    if (horizontal)
      return Coord(fp_start, fp_level);
    return Coord(fp_level, fp_start);
  };

  inline Coord end_pt() const {
    if (horizontal) 
      return Coord(fp_end, fp_level);
    return Coord(fp_level, fp_end);
  };

  friend class PixelSetBoundary;
  friend class PixelBdyLoop;
  friend std::ostream& operator<<(std::ostream &, const PixelBdySegment&);
};

std::ostream& operator<<(std::ostream &, const PixelBdySegment&);

class PixelBdyLoop {
private:
  std::vector<PixelBdySegment> loop;
  CRectangle *bounds;		// computed by set_floats
public:
  PixelBdyLoop() : bounds(0) {}
  ~PixelBdyLoop();
  void add_segment(const PixelBdySegment&);
  void clean();
  void set_floats(const Coord&);
  const CRectangle &bbox() const { return *bounds; }
  const std::vector<PixelBdySegment> &segments() const { return loop; }
};

typedef std::map<PixelBdySegment, int> SegMap;
typedef std::multimap<ICoord, PixelBdySegment> CoordMap;

// Pixel set boundary knows its microstructure.  The way this works
// is, you create it with a microstructure, then put in all the pixels
// in the pixel set for which you want this to be the boundary, then
// call the "find_boundary" method, which finds all the loops and
// assigns them to the "loopset" member.
class PixelSetBoundary {
private:
  SegMap segments;
  const CMicrostructure* microstructure;
  Coord ms_delta;
  std::vector<PixelBdyLoop*> loopset;
  PixelBdyLoop *find_loop(CoordMap&);
public:
  PixelSetBoundary(const CMicrostructure*);
  ~PixelSetBoundary();
  void add_pixel(const ICoord&);
  void find_boundary();
  const std::vector<PixelBdyLoop*> &get_loops() const { return loopset; }
  friend std::ostream& operator<<(std::ostream &, const PixelSetBoundary&);
};


// For debugging....
std::ostream& operator<<(std::ostream &, const PixelSetBoundary &);


#endif // PIXELSETBOUNDARY_H
