// -*- C++ -*-
// $RCSfile: cg.h,v $
// $Revision: 1.16 $
// $Author: vrc $
// $Date: 2007-09-26 22:03:10 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "common/tostring.h"
#include "common/IO/progressbar.h" 
#include "engine/log_value.h"
//*****************************************************************
// Iterative template routine -- CG
//
// CG solves the symmetric positive definite linear
// system Ax=b using the Conjugate Gradient method.
//
// CG follows the algorithm described on p. 15 in the 
// SIAM Templates book.
//
// The return value indicates convergence within max_iter (input)
// iterations (0), or no convergence within max_iter iterations (1).
//
// Upon successful return, output arguments have the following values:
//  
//        x  --  approximate solution to Ax = b
// max_iter  --  the number of iterations performed before the
//               tolerance was reached
//      tol  --  the residual after the final iteration
//  
//*****************************************************************

template < class Matrix, class Vector, class Preconditioner, class Real >
int 
CG(const Matrix &A, Vector &x, const Vector &b,
   const Preconditioner &M, int &max_iter, Real &tol)
{
  //Thread control starts here
  Progress * MyController = getProgress();
  std::string out_mssg;
  //Thread control ends here
  Real resid;
  Vector p, z, q;
  Vector alpha(1), beta(1), rho(1), rho_1(1);

  Real normb = norm(b);
  Vector r = b - A*x;

  if (normb == 0.0) 
    normb = 1;
  
  if ((resid = norm(r) / normb) <= tol) {
    //Thread control starts here
    out_mssg = "solving: " + to_string(resid) + "/" + to_string(tol);
    MyController->set_message(out_mssg);
    //MyController->set_success(); // notify that CG finished successfully (woo-hoo!)
    // tell the parent thread to stop monitoring and move on to next task
    //Thread control ends here
    tol = resid;
    max_iter = 0;
    
    
    return 0;
  }

  
  LogValue doubletolog(resid, tol);
  for (int i = 1; i <= max_iter && !MyController->query_stop(); i++) {
    z = M.solve(r);
    rho(0) = dot(r, z);
    
    if (i == 1)
      p = z;
    else {
      beta(0) = rho(0) / rho_1(0);
      p = z + beta(0) * p;
    }
    
    q = A*p;
    alpha(0) = rho(0) / dot(p, q);
    
    x += alpha(0) * p;
    r -= alpha(0) * q;
    
    if ((resid = norm(r) / normb) <= tol) {
      //Thread control starts here
      out_mssg = "solving: " + to_string(resid) + "/" + to_string(tol);
      MyController->set_message(out_mssg);
      //MyController->set_success(); // notify that CG finished successfully (woo-hoo!)
      //Thread control ends here
      tol = resid;
      max_iter = i;

      

      return 0;     
    }

    rho_1(0) = rho(0);
    
    //Thread control starts here
    out_mssg = "solving: " + to_string(resid) + "/" + to_string(tol);
    MyController->set_message(out_mssg);
    MyController->set_progress(doubletolog(resid)); 
    /* 
       Thread controller returns the current relative
       residual, which will be used to construct a 
       graphic or text bar.
    */
    //Thread control ends here
  }
  
  tol = resid;

  //Thread control starts here
  MyController->set_failure(); //If thread made it to this point, it means that solver failed.
  //Thread control ends here

  return 1;
}

